#!/usr/bin/env python3
"""Cluster tools
Usage:
  cluster_tools.py [-hv] [--register]
  cluster_tools.py [-hv] [--update]
  cluster_tools.py [-hv] [--deregister]
  cluster_tools.py --version
Options:
  -h --help            show this help message and exit
  --register           register with the cluster
  --update             check for new nodes to register with
  --deregister         de-register from the cluster
  --version            show version and exit
  -v --verbose         print status messages
"""


import socket
import time
import sys
import os
import re
from pathlib import Path
import subprocess as sp
import paramiko
from paramiko.client import SSHClient
from paramiko.sftp_client import SFTPClient
from paramiko.ssh_exception import NoValidConnectionsError
from docopt import docopt

UUID_PATTERN = re.compile(r"ipc_uuid *= *(.*)$")


class SSHCommandException(Exception):
    """ Raised on failure to execute a command over ssh """

    pass


class SSHTimeoutException(Exception):
    """ Raised on timeout of ssh action """

    pass


class RegistrationFailure(Exception):
    """ Raised when registration has failed and we should bail out """

    pass


class DeRegistrationFailure(Exception):
    """ Raised when de-registration has failed and we should bail out """

    pass


class UpdateFailure(Exception):
    """ Raised when a failure has occurded while updating the cluster configuration"""

    pass


class ShellExecutionFailure(Exception):
    """ Raised on failure to execute a shell cmd """

    pass


def info(*args):
    print("[>>>]", " ".join(map(str, args)))


def error(*args):
    print("[!!!]", " ".join(map(str, args)))


def execute_shell_cmd(cmd: list, error_msg: str) -> str:
    """Executes a shell command

    Args:
        cmd: A comma delimited list of strings which is the command to execute
        error_msg: String to print on error
    Raises:
        ShellExecutionFailure: On failure to execute the command
    Returns:
        The output of the execution
    """
    p = sp.Popen(
        cmd,
        stdin=sp.PIPE,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
    )
    output, err = p.communicate()
    if p.returncode != 0:
        error(error_msg)
        if VERBOSE:
            error(output.decode("utf-8"), err.decode("utf-8"))
        raise ShellExecutionFailure
    return output.decode("utf-8")


def get_uuid() -> str:
    """Gets the uuid from the merlin config

    If no uuid is found it generates a new one, adds it to the
    merlin config and then returns it.
    """
    with open(MERLIN_CONFIG, "r") as merlin_config:
        for line in merlin_config:
            match = UUID_PATTERN.match(line)
            if match:
                return match.group(1)

    # No UUID found
    uuid = generate_uuid()
    set_uuid(uuid)
    return uuid


def set_uuid(uuid: str):
    merlin_uuid = "ipc_uuid = {}\n".format(uuid)
    with open(MERLIN_CONFIG, "a") as file_handle:
        file_handle.write(merlin_uuid)


def generate_uuid() -> str:
    """ Generates and returns new UUID """

    try:
        output = execute_shell_cmd(
            ["mon", "id", "generate"], "Could not generate an UUID"
        )
    except ShellExecutionFailure:
        sys.exit(1)
    # Strip the newline
    return output.strip()


def parse_sync_files(sync_files: str) -> list:
    if sync_files:
        return [file_name.strip() for file_name in sync_files.split(",")]
    return []


VERBOSE = False
MERLIN_CONFIG = "/opt/monitor/op5/merlin/merlin.conf"
KNOWN_HOSTS = "/opt/monitor/.ssh/known_hosts"
PRIVATE_KEY = "/opt/monitor/.ssh/id_rsa"
DEFAULT_ENCRYPTION_KEYS_LOCATION = Path("/opt/monitor/op5/merlin/")
UUID = get_uuid()

try:
    POLLER_NAME = os.environ["POLLER_NAME"]
    MASTER_ADDRESS = os.environ["MASTER_ADDRESS"]
    MASTER_NAME = os.environ["MASTER_NAME"]
    MASTER_PORT = os.environ["MASTER_PORT"]
    POLLER_ADDRESS = os.environ["POLLER_ADDRESS"]
    POLLER_NAME = os.environ["POLLER_NAME"]
    POLLER_HOSTGROUP = os.environ["POLLER_HOSTGROUP"]
    POLLER_SSH_PORT = os.environ.get("POLLER_SSH_PORT", 22)
    CONNECT_TO_POLLERS = bool(os.environ.get("CONNECT_TO_POLLERS", 0))
    DISABLE_ENCRYPTION = bool(os.environ.get("DISABLE_ENCRYPTION", 0))
    FILES_TO_SYNC = parse_sync_files(os.environ.get("FILES_TO_SYNC"))
except KeyError as err:
    error("Required environment variable", err, "was not set, exiting")
    sys.exit(1)

def add_cluster_update() -> str:
    """ Adds cluster_update to merlin.conf """

    with open(MERLIN_CONFIG, "r") as merlin_config:
        for line in merlin_config:

            if "cluster_update" in line:
                return "cluster_update already set"

    try:
        output = execute_shell_cmd(
            ["sed", "-i", "/module {/a\\\tcluster_update = \/usr\/local\/bin\/cluster_tools.py --update", MERLIN_CONFIG], "Could not add cluster_update to merlin.conf"
        )
    except ShellExecutionFailure:
        raise RegistrationFailure
    # Strip the newline
    return output.strip()


def setup_ssh_connection(
    hostname: str, port: int = 22, timeout: float = 10.0
) -> SSHClient:
    """Sets up an SSH connection to a node

    Assumes that this node's public key has already
    been added to the remote node's authorized_keys.

    Will auto add the remote node to this node's
    known_hosts.

    Args:
        hostname: Hostname of the node
        port: SSH port of the node, default: 22
        timeout: Timeout in seconds for setting up the connection
    Raises:
        SSHTimeoutException: If timeout is reached
        NoValidConnectionsError: If connection can't be established
    Returns:
        SSHClient
    """
    client = SSHClient()
    try:
        client.load_host_keys(KNOWN_HOSTS)
    except FileNotFoundError:
        info(KNOWN_HOSTS, "not found, creating.")
        open(KNOWN_HOSTS, "w").close()
        client.load_host_keys(KNOWN_HOSTS)
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    try:
        client.connect(
            hostname=hostname, key_filename=PRIVATE_KEY, port=port, timeout=timeout
        )
    except socket.timeout:
        raise SSHTimeoutException
    return client


def execute_ssh_command(client: SSHClient, cmd: str, error_msg: str) -> str:
    """Executes a command over SSH

    Args:
        client: SSH-connection
        cmd: Command to execute
        error_msg: Message to print on failure
    Raises:
        SSHCommandException
    Returns:
        The output of the command if successful
    """
    _, stdout, stderr = client.exec_command(cmd)
    exit_code = stdout.channel.recv_exit_status()
    if exit_code != 0:
        error(error_msg)
        if VERBOSE:
            error("Error:", *stdout.readlines(), *stderr.readlines())
        raise SSHCommandException()
    else:
        output = stdout.readlines()
        # Strip whitespace and newlines
        return [item.strip() for item in output]


def acquire_master_lock(client: SSHClient):
    """Acquires a file lock on a master server

    This ensures only one poller can register at a time.

    Args:
        client: An ssh-connection to the server
    Returns:
        None
    Raises:
        SSHException: If the server fails to execute the command
    """
    # TODO: Add a file in created folder with node id for debugging
    cmd = 'mkdir "/tmp/poller_add.lock"'
    info("Attempting to acquire a lock on master")
    while True:
        _, stdout, stderr = client.exec_command(cmd)
        exit_code = stdout.channel.recv_exit_status()
        if exit_code == 0:
            break
        # TODO: Some kind of stale lock check?
        info("Failed to acquire lock, retrying in 5 seconds")
        if VERBOSE:
            error(" ".join(stderr.readlines()))
        time.sleep(5)
    info("Acquired master lock")


def release_master_lock(client: SSHClient):
    """Releases a file lock on a master server

    This ensures that only one poller registers at a time

    Args:
        client: An ssh-connection to the server
    Returns:
        None
    Raises:
        SSHException: If the server fails to execute the command
        SSHCommandException: If the command failed
    """
    cmd = 'rm -rf "/tmp/poller_add.lock"'
    _, stdout, stderr = client.exec_command(cmd)
    exit_code = stdout.channel.recv_exit_status()
    if exit_code == 0:
        info("Master lock released")
    else:
        info("Failed to release lock, might not be critical")
        if VERBOSE:
            error("Error:", " ".join(stderr.readlines()))


def add_master(master_name, master_address, poller_name, master_port=15551):
    """Add a config entry for a master node in the merlin config

    Args:
        master_name: Name of the new master
        master_address: Address of the new master
        poller_name: Name of this poller
        master_port: Port which merlin communicates over, default 15551
    Raises:
        pass
    """
    info("Adding", master_name, "as a master")
    sync_seperator = "\n    "
    if DISABLE_ENCRYPTION:
        encryption_block = "  encrypted = 0"
    else:
        encryption_block = (
            f"  encrypted = 1\n"
            f"  publickey = /opt/monitor/op5/merlin/{master_name}.pub"
        )
    config = (
        f"master {master_name} {{\n"
        f"  address = {master_address}\n"
        f"  port = {master_port}\n"
        f"{encryption_block}\n"
        f"  sync {{\n"
        f"    {sync_seperator.join(FILES_TO_SYNC)}\n"
        f"  }}\n"
        f"  object_config {{\n"
        f"    fetch_name = {poller_name}\n"
        f"    fetch = mon oconf fetch --sync {master_name}\n"
        f"  }}\n"
        f"}}\n"
    )

    with open(MERLIN_CONFIG, "a") as merlin_config:
        merlin_config.write(config)


def register_as_poller(client: SSHClient, name: str):
    """ Registers self as a poller on a master """
    info("Attempting to register as a poller on ", name)
    if not DISABLE_ENCRYPTION:
        public_key = exchange_encryption_keys(client, name)
    cmd = (
        "mon node add {} type=poller hostgroup={} "
        "address={} connect=no notifies=no uuid={}"
    ).format(POLLER_NAME, POLLER_HOSTGROUP, POLLER_ADDRESS, UUID)
    if not DISABLE_ENCRYPTION:
        cmd += f" publickey={public_key}"
    try:
        # Register poller
        execute_ssh_command(client, cmd, "Poller registration failed")
        # Restart master
        execute_ssh_command(
            client, "sudo mon restart", "Couldn't execute mon restart on the master"
        )
    except SSHCommandException:
        client.close()
        raise RegistrationFailure


def get_remote_cluster_info(client: SSHClient):
    """Retrives the cluster info from a node

    Args:
        client: SSH-connection to the node
    """
    info("Parsing remote cluster info")
    output = execute_ssh_command(
        client, "mon node show", "Could not retrive cluster info"
    )
    return parse_mon_node_show(output)


def get_local_cluster_info():
    """Retrives the local cluster info

    Args:
        client: SSH-connection to the node
    """
    info("Parsing local cluster info")
    try:
        output = execute_shell_cmd(
            ["mon", "node", "show"], "Could not fetch cluster info"
        )
    except ShellExecutionFailure:
        # Might not be an issue, will be caught later if it is
        return dict()
    if output == "No nodes configured\n":
        return dict()
    return parse_mon_node_show(output.split("\n"))


def parse_mon_node_show(cluster_info):
    """Parses the output of mon node show

    Args:
        cluster_info: The raw data as returned by mon node show
    Returns:
        parsed_info: A dict of the form {node: {var: val, ...}, ...}
    """
    parsed_info = {}
    current_node = None
    for line in cluster_info:
        if line in ["", "\n"]:
            continue
        if line.startswith("#"):
            # New node, looks like "# <name>\n"
            current_node = line[2:].strip()
            parsed_info[current_node] = {}
        else:
            # Normal peer info, "<VAR>=<val>\n"
            var, val = line.split("=")
            parsed_info[current_node][var.lower()] = val.strip()
    return parsed_info


def register_with_cluster(cluster_info, cross_register=False):
    """Goes through the cluster info and registers with every node

    Args:
        cross_register: If the poller should register with pollers as well.
    """
    for node in cluster_info:
        node_type = cluster_info[node]["type"]
        if node_type == "peer":
            master_info = cluster_info[node]
            client = setup_ssh_connection(master_info["address"])
            register_as_poller(client, master_info["name"])
            client.close()
            add_master(
                master_info["name"],
                master_info["address"],
                POLLER_NAME,
                master_info["port"],
            )
        elif (
            node_type == "poller"
            and cluster_info[node]["name"] != POLLER_NAME
            and cluster_info[node]["hostgroup"] == POLLER_HOSTGROUP
        ):
            add_poller(
                name=cluster_info[node]["name"], address=cluster_info[node]["address"]
            )
            if cross_register:
                register_with_poller(
                    name=cluster_info[node]["name"],
                    address=cluster_info[node]["address"],
                )


def add_poller(name: str, address: str):
    """Add a poller entry to the merlin config

    Args:
        name: Name of the new poller
        address: Address to the new poller
    """
    info("Adding", name, "as a poller-peer")
    try:
        execute_shell_cmd(
            ["mon", "node", "add", name, "type=peer", "address=" + address],
            "Failed to add poller-peer",
        )
    except ShellExecutionFailure:
        raise RegistrationFailure


def register_with_poller(name: str, address: str):
    """Registers with a poller
    Args:
        name: Name of the new poller
        address: Address to the new poller
    Note: This requires the poller to have an ssh-daemon
    """
    client = setup_ssh_connection(address, port=POLLER_SSH_PORT)
    add_cmd = "mon node add {} type=peer address={} connect=no".format(
        POLLER_NAME, POLLER_ADDRESS
    )
    try:
        execute_ssh_command(client, add_cmd, "Failed to register with poller: " + name)
    except SSHCommandException:
        raise RegistrationFailure
    finally:
        client.close()


def restart_cluster():
    info("Executing mon restart on the whole cluster")
    try:
        execute_shell_cmd(
            ["mon", "node", "ctrl", "--self", "--", "sudo mon restart"],
            "Failed to execute restart",
        )
    except ShellExecutionFailure:
        pass


def fetch_config():
    """ Fetches the naemon config """
    info("Fetching config from", MASTER_NAME)
    try:
        execute_shell_cmd(
            ["mon", "oconf", "fetch", "--sync", MASTER_NAME],
            "Could not fetch configuration",
        )
    except ShellExecutionFailure:
        raise RegistrationFailure


def deregister_with_cluster():
    """Connects too all masters and removes self as poller
    Args:
        deregister_from_pollers
    """
    info("Exiting cluster")
    cluster_info = get_local_cluster_info()
    for node in cluster_info:
        if cluster_info[node]["type"] == "master":
            deregister_from_master(
                name=cluster_info[node]["name"], address=cluster_info[node]["address"]
            )
        elif (
            CONNECT_TO_POLLERS
            and cluster_info[node]["type"] == "peer"
            and cluster_info[node]["name"] != POLLER_NAME
        ):
            deregister_from_poller(
                name=cluster_info[node]["name"],
                address=cluster_info[node]["address"],
            )
    clean_up_local_state()


def deregister_from_master(name: str, address: str):
    """Connects to a master and removes self as poller

    Args:
        name: Name of master
        address: Address of the master
    """
    info("De-registering from master:", name)
    client = setup_ssh_connection(address)
    try:
        if not DISABLE_ENCRYPTION:
            # Remove public encryption key
            cluster_info = get_remote_cluster_info(client)
            if cluster_info.get(POLLER_NAME, dict()).get("publickey"):
                execute_ssh_command(
                    client,
                    f'rm -f {cluster_info[POLLER_NAME]["publickey"]}',
                    "Could not remove public encryption key from {name}",
                )
        # Remove self
        execute_ssh_command(
            client, "mon node remove " + POLLER_NAME, "Peer de-registration failed"
        )
        # Restart peer
        execute_ssh_command(
            client, "sudo mon restart", "Failed to restart master: " + name
        )
    except SSHCommandException:
        raise DeRegistrationFailure
    finally:
        client.close()


def remove_node(name: str):
    """Removes a node from the configuration

    Args:
        name: Name of the node to be removed
    """
    info("Removing node:", name)
    try:
        execute_shell_cmd(["mon", "node", "remove", name], "Could not remove node")
        execute_shell_cmd(
            ["rm", "-f", f"/opt/monitor/op5/merlin/{name}.pub"],
            f"Could not remove public encryption key belonging to {name}",
        )
    except ShellExecutionFailure:
        pass


def deregister_from_poller(name: str, address: str):
    """Connects to a poller and removes self as peer

    Args:
        name: Name of the poller
        address: Address of the poller
    """
    info("De-registering from poller:", name)
    client = setup_ssh_connection(address)
    try:
        # Remove self
        execute_ssh_command(
            client, "mon node remove " + POLLER_NAME, "poller de-registration failed"
        )
        # Restart poller
        execute_ssh_command(
            client, "sudo mon restart", "Failed to restart poller: " + name
        )
    except SSHCommandException:
        raise DeRegistrationFailure
    finally:
        client.close()


def hostgroup_exists(client: SSHClient, hostgroup) -> bool:
    """Queries livestatus on a node to check if a hostgroups exists

    Args:
        client: SSH-connection to the node
        hostgroup: Name of the hostgroup to check
    Returns:
        True, if the hostgroup exists
    """

    info("Verifying that hostgroup", hostgroup, "exists on the master")
    cmd = "mon query ls hostgroups -c name name={}".format(hostgroup)
    try:
        output = execute_ssh_command(
            client, cmd, "Something went wrong when querying livestatus"
        )
    except SSHCommandException:
        return False
    return hostgroup in output


def restart():
    """ Runs mon restart locally """
    info("Executing mon restart")
    try:
        execute_shell_cmd(["mon", "restart"], "Could not restart")
    except ShellExecutionFailure:
        pass


def update():
    """Updates the cluster configuration

        Contacts the master and checks for new nodes and nodes which have been removed

    Raises:
        UpdateFailure: On failure to establish a connection to the master.
    """
    local_cluster_info = get_local_cluster_info()
    # Add our selves to the configuration so we aren't flagged as new
    local_cluster_info[POLLER_NAME] = {}
    # Get remote configuration
    try:
        client = setup_ssh_connection(MASTER_ADDRESS)
    except (SSHTimeoutException, NoValidConnectionsError):
        # Master is unreachable
        raise UpdateFailure("Can't establish connection to master")
    remote_cluster_info = get_remote_cluster_info(client)
    # Add the master to it's own config so it isn't flagged for removal
    remote_cluster_info[MASTER_NAME] = {}
    client.close()

    # Nodes to add
    for node in remote_cluster_info.keys() - local_cluster_info.keys():
        node_info = remote_cluster_info[node]
        if node_info["type"] == "peer":
            info("New master:", node_info["name"])
            add_master(
                node_info["name"],
                node_info["address"],
                POLLER_NAME,
                node_info["port"],
            )
            try:
                client = setup_ssh_connection(node_info["address"])
                register_as_poller(client=client, name=node_info["name"])
            except SSHTimeoutException:
                error("SSH connection to", node_info["name"], "timed out")
            except RegistrationFailure:
                error("Registration failed, manual intervention needed")
            finally:
                client.close()
        elif (
            node_info["type"] == "poller" and node_info["hostgroup"] == POLLER_HOSTGROUP
        ):
            info("New poller:", node_info["name"])
            add_poller(name=node_info["name"], address=node_info["address"])

    # Nodes to remove
    for node in local_cluster_info.keys() - remote_cluster_info.keys():
        remove_node(local_cluster_info[node]["name"])
    restart()


def exchange_encryption_keys(client: SSHClient, name: str) -> str:
    """Exchanges public encryption keys with another node

    Args:
        client: SSHClient connection the the remote node
        name: Name of the remote node
    Returns:
        Path to this node's public key on the remote node
    Raises:
        RegistrationFailure: On failure to exchange keys
    """
    info("Exchanging keys with", name)
    # Location of encryption keys on remote node
    encryption_keys_location = DEFAULT_ENCRYPTION_KEYS_LOCATION
    try:
        # Check if path to private key has been set
        output = execute_ssh_command(
            client,
            f"awk '/ipc_privatekey/{{print $NF}}' {MERLIN_CONFIG}",
            "Failed to awk the merlin config for location of private key",
        )
        if output:
            info("Found path to private key in merlin config on", name)
            # Output will be a list with one item
            encryption_keys_location = Path(output.pop()).parent
        else:
            generate_encryption_keys_on_remote(client, name)
    except SSHCommandException:
        raise RegistrationFailure
    try:
        with SFTPClient.from_transport(client.get_transport()) as sftp:
            sftp.get(
                str(encryption_keys_location / "key.pub"),
                str(DEFAULT_ENCRYPTION_KEYS_LOCATION / f"{name}.pub"),
            )
            sftp.put(
                str(DEFAULT_ENCRYPTION_KEYS_LOCATION / "key.pub"),
                str(encryption_keys_location / f"{POLLER_NAME}.pub"),
            )
    except FileNotFoundError:
        error("Encryption keys not found on remote master", name)
        raise RegistrationFailure
    return str(encryption_keys_location / f"{POLLER_NAME}.pub")


def generate_encryption_keys_on_remote(client: SSHClient, name: str):
    """Generates encryption keys on a remote node
    Also adds the path to the private key to the merlin config.

    Args:
        client: SSHClient connection to the remote node
        name: Name of the remote node
    Raises:
        RegistrationFailure
    """
    info("Generating encryption keys on", name)
    try:
        # Generate encryption-key-pair
        execute_ssh_command(
            client,
            "mon merlinkey generate",
            f"Could not generate encryption keys on {name}",
        )
    except SSHCommandException:
        # Check the rare case where keys exists on the host
        # at the default location but the path to private is
        # not written to the merlin config.
        try:
            pub_key_location = str(DEFAULT_ENCRYPTION_KEYS_LOCATION / "key.pub")
            priv_key_location = str(DEFAULT_ENCRYPTION_KEYS_LOCATION / "key.priv")
            execute_ssh_command(
                client,
                f"test -e {pub_key_location} && test -e {priv_key_location}",
                f"One or both of the encryption keys were not found at {DEFAULT_ENCRYPTION_KEYS_LOCATION}",
            )
            error(
                "Encryption key-pair found on host but location of private key was not written to the merlin configuration file"
            )
            info("Correcting this error")
        except SSHCommandException:
            raise RegistrationFailure
    try:
        # Write path to private key to config
        execute_ssh_command(
            client,
            f'echo "ipc_privatekey = /opt/monitor/op5/merlin/key.priv" >> {MERLIN_CONFIG}',
            f"Could not write location of private key to the merlin config on {name}",
        )
    except SSHCommandException:
        raise RegistrationFailure


def generate_encryption_keys():
    """Generates local encryption keys

    Raises:
        RegistrationFailure
    """
    info("Generating local encryption keys")
    # Check if keys already exist
    priv_key_exists = (DEFAULT_ENCRYPTION_KEYS_LOCATION / "key.priv").exists()
    pub_key_exists = (DEFAULT_ENCRYPTION_KEYS_LOCATION / "key.pub").exists()
    if pub_key_exists and priv_key_exists:
        # No need to generate encryption keys
        return
    if priv_key_exists or pub_key_exists:
        error(
            "Only one of keys in encryption key pair exists, manual intervention needed"
        )
        raise RegistrationFailure
    try:
        execute_shell_cmd(
            ["mon", "merlinkey", "generate"],
            "Could not generate encryption keys",
        )
    except ShellExecutionFailure:
        raise RegistrationFailure
    # Write location of private key to merlin config
    with open(MERLIN_CONFIG, "a") as fh:
        fh.write("ipc_privatekey = /opt/monitor/op5/merlin/key.priv\n")


def clean_up_local_state():
    """ Clean up local state """
    info("Cleaning up local state")
    cmd = ["mon", "node", "remove"]
    local_info = get_local_cluster_info()
    if not local_info:
        return
    cmd.extend(local_info.keys())
    try:
        execute_shell_cmd(cmd, "Could not remove local node configuration")
    except ShellExecutionFailure:
        raise DeRegistrationFailure


def main():
    arguments = docopt(__doc__, version="1.0")
    if arguments.get("-v") or arguments.get("--verbose"):
        global VERBOSE
        VERBOSE = True
    if arguments.get("--deregister"):
        try:
            master = setup_ssh_connection(MASTER_ADDRESS)
            acquire_master_lock(master)
            deregister_with_cluster()
        except DeRegistrationFailure:
            error("De-registration failed, manual intervention needed")
            sys.exit(1)
        finally:
            release_master_lock(master)
            master.close()
    elif arguments.get("--update"):
        try:
            update()
        except UpdateFailure as ex:
            error(ex)
    elif arguments.get("--register"):
        if MASTER_NAME in get_local_cluster_info():
            info("It appears this node is already configured, no registration needed.")
            sys.exit(0)
        master = setup_ssh_connection(MASTER_ADDRESS)
        if not hostgroup_exists(master, POLLER_HOSTGROUP):
            error("Hostgroup does not exist")
            sys.exit(1)
        try:
            if not DISABLE_ENCRYPTION:
                generate_encryption_keys()
            acquire_master_lock(master)
            # Register with master
            add_master(MASTER_NAME, MASTER_ADDRESS, POLLER_NAME)
            register_as_poller(client=master, name=MASTER_NAME)
            cluster_info = get_remote_cluster_info(master)
            fetch_config()
            # Register with the rest of the cluster
            register_with_cluster(cluster_info, cross_register=CONNECT_TO_POLLERS)
            release_master_lock(master)
            master.close()
            add_cluster_update()
            restart_cluster()
        except RegistrationFailure:
            # Bail and clean up as much as possible
            error("Error occurred during registration, bailing out and cleaning up")
            release_master_lock(master)
            master.close()
            deregister_with_cluster()
            sys.exit(1)
        # Sanity check that the poller is now added
        if MASTER_NAME not in get_local_cluster_info():
            error(
                "Something went wrong during setup.",
                MASTER_NAME,
                "not in local configuration",
            )
        else:
            info("Poller", POLLER_NAME, "successfully registered with cluster")
    else:
        info("Nothing to do")


if __name__ == "__main__":
    main()
