#!/bin/bash

function show_usage {
  printf '%s\n' "[Description]"
  printf '  %s\n' "Command line utility to setup an OP5 Slim Poller with an OP5 master"

  printf '\n%s\n' "[Usage]"
  printf '  %s\n' "$0 --master-name <MASTER_NAME> --master-address <MASTER_IP> --poller-name <POLLER_NAME>"

  printf '\n%s\n' "[Flags]"
  printf '  %s\t%s\n' "--master-name"     "Name of the OP5 master [Required]"
  printf '  %s\t%s\n' "--master-address"  "IP/Hostname of the OP5 master [Required]"
  printf '  %s\t%s\n' "--master-port"	  "Port to use for merlin communication (default: 15551)"
  printf '  %s\t%s\n' "--poller-name"	  "Name of the poller [Required]"
}

# Handle cmd line arguments
master_name=
master_address=
master_port=15551
poller_name=

options=$(getopt -l "help,master-name:,master-address:,master-port:,poller-name:" -o "hn:i:p:N:" -a -- "$@")

eval set -- "$options"

while true
do
case $1 in
-h|--help)
  show_usage
  exit 0
  ;;
-n|--master-name)
  shift
  master_name=$1
  ;;
-i|--master-address)
  shift
  master_address=$1
  ;;
-p|--master-port)
  shift
  master_port=$1
  ;;
-N|--poller-name)
  shift
  poller_name=$1
  ;;
--)
  shift
  break;;
esac
shift
done

# make sure all required commandline arguments are supplied
argument_missing=0
if [[ -z "$master_name" ]]; then
  echo "[>>>] Argument master-name is required"
  argument_missing=1
fi
if [[ -z "$master_address" ]]; then
  echo "[>>>] Argument master-address is required"
  argument_missing=1
fi
if [[ -z "$poller_name" ]]; then
  echo "[>>>] Argument poller-name is required"
  argument_missing=1
fi

if [ $argument_missing -eq 1 ]; then
  echo ""
  show_usage
  exit 1
fi

# Make sure poller_name has not yet been added to merlin
node_info=$(mon node show $master_name)

if [ $? -ne 0 ]; then
  echo "$node_info"
  echo "[>>>]"
  echo "[>>>] Error:"
  echo "[>>>] Something went wrong while checking current nodes."
fi

if [[ $node_info == *"NAME=$master_name"* ]]; then
  echo "[>>>]"
  echo "[>>>] Error:"
  echo "[>>>] It appears this node is already configured. Aborting..."
  exit 1
fi

# Check SSH connection
echo "[>>>] Checking SSH connection"
ssh -q -o PasswordAuthentication=no -o StrictHostKeyChecking=no $master_address exit
ssh_ok=$?

if [ $ssh_ok -ne 0 ]; then
  echo "[>>>]"
  echo "[>>>] Error:"
  echo "[>>>] SSH connection could not be established."
  echo "[>>>] Make sure SSH keys have been synced, that the IP is correct, and that there are no firewall issues."
  exit 1
fi


# Verify master has setup the poller
echo "[>>>] Verifying $master_name configuration"
ssh -q $master_address "test -e /var/cache/merlin/config/$poller_name.cfg"
if [ $? -ne 0 ]; then
  echo "[>>>]"
  echo "[>>>] Error:"
  echo "[>>>] It appears the master server has not been correctly setup."
  echo "[>>>] Ensure that it has a node: $poller_name, and you have executed mon restart"
  exit 1
fi

# Add merlin node
echo "[>>>] Adding node: $master_name"
cat <<EOF >> /opt/monitor/op5/merlin/merlin.conf
master $master_name {
  address = $master_address
  port = $master_port
  object_config {
    fetch_name = $poller_name
    fetch = mon oconf fetch $master_name
  }
}
EOF

if [ $? -ne 0 ]; then
  echo "[>>>]"
  echo "[>>>] Error:"
  echo "[>>>] Could not add the master node to the merlin configuration"
  exit 1
fi

# fetch config
echo "[>>>] Fetching config from $master_name"
fetch_output="$(mon oconf fetch $master_name 2>&1)"

if [ $? -ne 0 ]; then
  echo "[>>>]"
  echo "[>>>] Error:"
  echo "[>>>] Could not fetch configuration from $master_name"
  echo "[>>>] Logs:"
  echo ""
  echo "$fetch_output"
  exit 1
fi

# mon oconf fetch does not always restart. For example when the current
# configuration is valid, but we've just added a new master node (i.e the
# second master of a peered master system). To ensure we always restart after
# adding a node, we do mon restart manually here.
mon restart

# Sanity check that the poller_name is now added
node_info=$(mon node show $master_name)

if [ $? -ne 0 ]; then
  echo "$node_info"
  echo "[>>>]"
  echo "[>>>] Error:"
  echo "[>>>] Something went wrong while checking current nodes."
fi

if [[ $node_info != *"NAME=$master_name"* ]]; then
  echo "[>>>]"
  echo "[>>>] Error:"
  echo "[>>>] $master_name does not appear to be added correctly"
  exit 1
fi

# Everything seems ok
echo "[>>>] Master $master_name with address: $master_address was successfully setup"
